// Copyright 2017 Microsoft Corporation. All rights reserved.
// Use of this source code is governed by an MIT
// license that can be found in the LICENSE file.

// Package policy contains the definitions needed for configuring in-box pipeline policies
// and creating custom policies.
package policy
