// Copyright 2017 Microsoft Corporation. All rights reserved.
// Use of this source code is governed by an MIT
// license that can be found in the LICENSE file.

// Package streaming contains helpers for streaming IO operations and progress reporting.
package streaming
